from typing import (
    get_type_hints,
    get_origin,
    get_args,
    Union,
    Any,
    Callable,
    Optional,
)
import inspect


def check_types(warnings: bool = True, strictfloat: bool = False):
    """
    A decorator that checks the types of the arguments passed to a function.
    It raises a TypeError if the type of an argument is not compatible with the type hint.

    Parameters:
    -----------
    warnings : bool, optional
        Whether to print warnings when a type cannot be checked, by default True
    strictfloat : bool, optional
        If True, ints will raise an error if a float is expected, by default False, in which case ints will be converted to floats
    """

    def decorator(func: Callable):
        def wrapper(*args, **kwargs):
            ### get type hints and variable names
            hints = get_type_hints(func)
            var_names = list(inspect.signature(func).parameters.keys())

            args = list(args)

            ### check args
            for arg_idx, arg in enumerate(args):
                arg_name = var_names[arg_idx]
                if arg_name in hints.keys():
                    check_type_hint(
                        var_name=arg_name,
                        var=args,
                        var_idx=arg_idx,
                        type_hint=hints[arg_name],
                        func=func,
                        warnings=warnings,
                        strictfloat=strictfloat,
                    )

            ### check kwargs
            for kwarg_name, kwarg in kwargs.items():
                check_type_hint(
                    var_name=kwarg_name,
                    var=kwargs,
                    var_idx=kwarg_name,
                    type_hint=hints[kwarg_name],
                    func=func,
                    warnings=warnings,
                    strictfloat=strictfloat,
                )
            return func(*args, **kwargs)

        return wrapper

    return decorator


def check_class_base_rec(var_name, func, type_hint, args, base):
    ### check if args of type hint is equal to base, if yes, return None
    if not args == base:
        ### if not equal go deeper with base
        if base == None:
            ### if base already None, raise TypeError
            raise TypeError(
                f"Parameter {var_name} of function {func} should be a class of {get_args(type_hint)[0]}"
            )
        else:
            ### go deeper and check base of base of base
            return check_class_base_rec(var_name, func, type_hint, args, base.__base__)
    else:
        return None


def check_type_hint(
    var_name: str,
    var: Any,
    var_idx: Any,
    type_hint: type,
    func: Callable,
    warnings: bool,
    strictfloat: bool = False,
    current_var: Optional[Any] = None,
    current_type_hint: Optional[type] = None,
) -> None:
    """
    Check if the type of the variable is compatible with the type hint

    Parameters
    ----------
    var_name : str
        Name of the variable
    var : Any
        List or dict containing variable to check
    var_idx : Any
        Index or key of the variable to check
    type_hint : type
        Type hint of the variable
    func : function
        Function that is checked
    warnings : bool, optional
        Whether to print warnings when a type cannot be checked, by default True
    strictfloat : bool, optional
        If True, ints will raise an error if a float is expected, by default False, in which case ints will be converted to floats
    current_var : Any, optional
        List or dict containing current variable to check, used for recursive calls, by default None
    current_type_hint : type, optional
        Current type hint of the variable, used for recursive calls, by default None
    """

    ### set current_var and current_type_hint to var and type_hint if they are None
    if current_var is None:
        current_var = var
    if current_type_hint is None:
        current_type_hint = type_hint

    # print(f"var_name: {var_name}, var: {current_var}, var_idx: {var_idx}")
    # print(
    #     f"var_name: {var_name}, var: {current_var}, var_type: {type(current_var)}, current_type_hint: {current_type_hint}, origin: {get_origin(current_type_hint)}, args: {get_args(current_type_hint)}"
    # )

    ### depending on origin of type hint doe different things
    ### if origin is type
    ### check if args of type hint is equal to var
    ### if not, check if equal to base of var recursively
    ### if never true, raise TypeError
    if get_origin(current_type_hint) is type:
        ### Type expects a class, first check if var is a class
        if not inspect.isclass(current_var[var_idx]):
            raise TypeError(
                f"Parameter {var_name} of function {func} should be a class of {get_args(type_hint)[0]}"
            )
        ### check if args of type hint is equal to var, if yes, return None
        if get_args(current_type_hint)[0] == current_var[var_idx]:
            return None
        else:
            ### if not equal, check if base class is equal to args of type hint
            return check_class_base_rec(
                var_name,
                func,
                type_hint,
                args=get_args(current_type_hint)[0],
                base=current_var[var_idx].__base__,
            )

    ### if origin is dict
    ### check if type is dict
    ### check if type of all keys and vals are correct
    ### do this by calling function check_type_hint recursively
    if get_origin(current_type_hint) is dict:
        ### check if type is dict
        if not isinstance(current_var[var_idx], dict):
            raise TypeError(
                f"Parameter {var_name} of function {func} should be {type_hint}"
            )
        ### passed dict check, now check if type of keys and vals are correct
        type_key = get_args(current_type_hint)[0]
        type_val = get_args(current_type_hint)[1]
        for key, val in current_var[var_idx].items():
            check_type_hint(
                var_name=var_name,
                var=var,
                var_idx=list(current_var[var_idx].keys()).index(key),
                type_hint=type_hint,
                func=func,
                warnings=warnings,
                strictfloat=strictfloat,
                current_var=list(current_var[var_idx].keys()),
                current_type_hint=type_key,
            )
            check_type_hint(
                var_name=var_name,
                var=var,
                var_idx=key,
                type_hint=type_hint,
                func=func,
                warnings=warnings,
                strictfloat=strictfloat,
                current_var=current_var[var_idx],
                current_type_hint=type_val,
            )
        ### variable passed all checks, return None
        return None

    ### if origin is list
    ### check if type is list
    ### check if type of all elements is correct
    ### do this by calling function check_type_hint recursively
    elif get_origin(current_type_hint) is list:
        ### check if type is list
        if not isinstance(current_var[var_idx], list):
            raise TypeError(
                f"Parameter {var_name} of function {func} should be {type_hint}"
            )
        ### passed list check, now check if type of all elements is correct
        type_element = get_args(current_type_hint)[0]
        for element_idx in range(len(current_var[var_idx])):
            check_type_hint(
                var_name=var_name,
                var=var,
                var_idx=element_idx,
                type_hint=type_hint,
                func=func,
                warnings=warnings,
                strictfloat=strictfloat,
                current_var=current_var[var_idx],
                current_type_hint=type_element,
            )
        ### variable passed all checks, return None
        return None

    ### if origin is Union
    ### check if variable is one of the types in the union
    ### do this by calling function check_type_hint recursively
    elif get_origin(current_type_hint) is Union or get_origin(
        current_type_hint
    ) is type(str | int):
        ### check if variable is any of the types in the union
        for type_ in get_args(current_type_hint):
            ### need a try except here because we want to check all types in the union and not stop at the first one that fails
            try:
                check_type_hint(
                    var_name=var_name,
                    var=var,
                    var_idx=var_idx,
                    type_hint=type_hint,
                    func=func,
                    warnings=warnings,
                    strictfloat=strictfloat,
                    current_var=current_var,
                    current_type_hint=type_,
                )
                ### variable passed one check, return None
                return None
            except:
                pass
        ### variable did not pass any checks, raise TypeError
        raise TypeError(
            f"Parameter {var_name} of function {func} should be {type_hint}"
        )

    ### if origin is None, its a simple type and we can check it directly using isinstance
    elif get_origin(current_type_hint) is None:
        ### check if variable is an instance of the type hint, if not raise TypeError
        ### catch if float type is expected as type hint but the type of the variable is int
        ### this should not raise an error ints can simply be converted to float (nothing is lost, in contrast to float->int)
        if not isinstance(current_var[var_idx], current_type_hint):
            if (
                current_type_hint is float
                and isinstance(current_var[var_idx], int)
                and not strictfloat
            ):
                current_var[var_idx] = float(current_var[var_idx])
                return None
            raise TypeError(
                f"Parameter {var_name} of function {func} should be {type_hint}"
            )
        else:
            return None

    ### if variable passed all checks without errors or returning, print that the type cannot be checked
    if warnings:
        print(
            f"WARNING check_type_hint: Variable {var_name} of function {func} could not be checked."
        )
    return None



from typingchecker import check_types
import pytest


class checked_class:
    @check_types()
    def __init__(
        self,
        a: int,
        b: float,
        c: list[float] = [1.0, 2.0, 3.0],
        d: dict[str, float] = {"a": 1.0, "b": 2.0, "c": 3.0},
    ):
        self.a = a
        self.b = b
        self.c = c
        self.d = d


def test_check_types_convert_int():
    ####################################################################################
    ############################   RAISING NO ERRORS   #################################
    ####################################################################################

    ### give an int to b, this should be converted to float
    checked_obj = checked_class(1, 2)
    assert isinstance(checked_obj.a, int) and isinstance(
        checked_obj.b, float
    ), "check_types did not convert int to float"

    ### give a list of ints to c, this should be converted to list of floats
    checked_obj = checked_class(1, 2.0, [1, 6.0, 3])
    assert (
        isinstance(checked_obj.a, int)
        and isinstance(checked_obj.b, float)
        and isinstance(checked_obj.c[0], float)
    ), "check_types did not convert list of int to list of float"

    ### give a dict with ints to d, this should be converted to dict with floats
    checked_obj = checked_class(1, 2.0, [1, 6.0, 3], {"a": 1, "b": 2.0, "c": 3})
    assert (
        isinstance(checked_obj.a, int)
        and isinstance(checked_obj.b, float)
        and isinstance(checked_obj.c[0], float)
        and isinstance(checked_obj.d["a"], float)
    ), "check_types did not convert list of int to list of float"

    ####################################################################################
    ##############################   RAISING ERRORS   ##################################
    ####################################################################################

    ### keys should still be strings
    with pytest.raises(TypeError):
        checked_obj = checked_class(1, 2.0, [1, 6.0, 3], {1: 1, "b": 2.0, "c": 3})

    ### floats instead of ints should still raise an error
    with pytest.raises(TypeError):
        checked_obj = checked_class(1.5, 2.0, [1, 6.0, 3], {"a": 1, "b": 2.0, "c": 3})


if __name__ == "__main__":
    test_check_types_convert_int()



from typingchecker import check_types
from typing import Union, Type
import pytest


class needed_class:
    def __init__(self) -> None:
        pass


### subclass of object, actually the same as needed_class, but for testing...
class needed_obj_class(object):
    def __init__(self) -> None:
        pass


### subclass of int
class needed_int_class(int):
    def __init__(self) -> None:
        pass


class checked_class:
    @check_types()
    def __init__(
        self,
        a: int,
        b: list[list[int]],
        c: Union[needed_class, list[needed_class]],
        d: float | list[float],
        e: dict[str, int],
        # f: Type expects a subclass of the specified class, not an instance
        f: Type[needed_class] = needed_class,
        g: Type[needed_obj_class] = needed_obj_class,
        h: Type[needed_int_class] = needed_int_class,
        # i: this expects a subclass of the object class which should be any class
        i: Type[object] = object,
        j: Type[needed_int_class] | Type[needed_obj_class] = needed_int_class,
    ):
        pass


needed_obj = needed_class()


def test_check_types_decorator():
    ####################################################################################
    ############################   RAISING NO ERRORS   #################################
    ####################################################################################

    ### if check_types works, this should not raise an error
    checked_class(
        1,
        [[1]],
        needed_obj,
        1.0,
        {"a": 1},
    )

    ### if check_types works, this should not raise an error
    ### d expects float, but int can simply be used as float...
    checked_class(
        1,
        [[1]],
        needed_obj,
        1,
        {"a": 1},
    )

    ### if check_types works, this should not raise an error
    ### d expects float, but int can simply be used as float...
    checked_class(
        1,
        [[1]],
        needed_obj,
        [1],
        {"a": 1},
    )
    ### if check_types works, this should not raise an error
    ### d expects float, but int can simply be used as float...
    checked_class(
        1,
        [[1]],
        needed_obj,
        [1, 2],
        {"a": 1},
    )

    ### if check_types works, this should not raise an error
    checked_class(
        1,
        [[1], [2]],
        [needed_obj, needed_obj],
        [1.0, 2.0],
        {"a": 1, "b": 2},
    )

    ### if check_types works, this should not raise an error
    checked_class(
        1,
        [],
        needed_obj,
        1.0,
        {},
    )

    ### if check_types works, this should not raise an error
    checked_class(
        1,
        [[]],
        [needed_obj],
        [1.0],
        {"a": 1},
    )

    ### if check_types works, this should not raise an error
    checked_class(
        1,
        [[1]],
        needed_obj,
        1.0,
        {"a": 1},
        f=needed_class,
        g=needed_obj_class,
        h=needed_int_class,
        i=object,
        j=needed_int_class,
    )

    ### if check_types works, this should not raise an error
    checked_class(
        1,
        [[1]],
        needed_obj,
        1.0,
        {"a": 1},
        f=needed_class,
        g=needed_obj_class,
        h=needed_int_class,
        i=object,
        j=needed_obj_class,
    )

    ### if check_types works, this should not raise an error
    checked_class(
        1,
        [[1]],
        needed_obj,
        1.0,
        {"a": 1},
        f=needed_class,
        g=needed_obj_class,
        h=needed_int_class,
        i=needed_obj_class,
        j=needed_obj_class,
    )

    ### if check_types works, this should not raise an error
    checked_class(
        1,
        [[1]],
        needed_obj,
        1.0,
        {"a": 1},
        f=needed_class,
        g=needed_obj_class,
        h=needed_int_class,
        i=needed_int_class,
        j=needed_obj_class,
    )

    ### if check_types works, this should not raise an error
    checked_class(
        1,
        [[1]],
        needed_obj,
        1.0,
        {"a": 1},
        f=needed_class,
        g=needed_obj_class,
        h=needed_int_class,
        i=needed_class,
        j=needed_obj_class,
    )

    ####################################################################################
    #############################   RAISING ERRORS   ###################################
    ####################################################################################

    ### if check_types works, this should raise an TypeError because a is not an int
    with pytest.raises(TypeError):
        checked_class(
            "1",
            [[1]],
            needed_obj,
            1.0,
            {"a": 1},
        )

    ### if check_types works, this should raise an TypeError because a is not an int
    with pytest.raises(TypeError):
        checked_class(
            1.2,
            [[1]],
            needed_obj,
            1.0,
            {"a": 1},
        )

    ### if check_types works, this should raise an TypeError because b is not a list of lists of ints
    with pytest.raises(TypeError):
        checked_class(
            1,
            [1],
            needed_obj,
            1.0,
            {"a": 1},
        )

    ### if check_types works, this should raise an TypeError because c is not a needed_class or a list of needed_class
    with pytest.raises(TypeError):
        checked_class(
            1,
            [[1]],
            "needed_obj",
            1.0,
            {"a": 1},
        )

    ### if check_types works, this should raise an TypeError because d is not a float or a list of floats
    with pytest.raises(TypeError):
        checked_class(
            1,
            [[1]],
            needed_obj,
            "1.0",
            {"a": 1},
        )

    ### if check_types works, this should raise an TypeError because e is not a dict of str to int
    with pytest.raises(TypeError):
        checked_class(
            1,
            [[1]],
            needed_obj,
            1.0,
            {"a": "1"},
        )

    ### if check_types works, this should raise an TypeError because c is not an instance of needed_class but the class itself
    with pytest.raises(TypeError):
        checked_class(
            1,
            [[1]],
            needed_class,
            1.0,
            {"a": 1},
        )

    ### if check_types works, this should raise an TypeError because f is not the expected needed_class but a instance of needed_class
    with pytest.raises(TypeError):
        checked_class(
            1,
            [[1]],
            needed_obj,
            1.0,
            {"a": 1},
            f=needed_obj,
        )

    ### if check_types works, this should raise an TypeError because h is not the expected needed_int_class but an int instance
    with pytest.raises(TypeError):
        checked_class(
            1,
            [[1]],
            needed_obj,
            1.0,
            {"a": 1},
            h=2,
        )

    ### if check_types works, this should raise an TypeError because f epxects a specific class and object is too general
    with pytest.raises(TypeError):
        checked_class(
            1,
            [[1]],
            needed_obj,
            1.0,
            {"a": 1},
            f=object,
        )

    ### if check_types works, this should raise an TypeError because g epxects a specific class and object is too general
    with pytest.raises(TypeError):
        checked_class(
            1,
            [[1]],
            needed_obj,
            1.0,
            {"a": 1},
            g=object,
        )

    ### if check_types works, this should raise an TypeError because h epxects a specific class and object is too general
    with pytest.raises(TypeError):
        checked_class(
            1,
            [[1]],
            needed_obj,
            1.0,
            {"a": 1},
            h=object,
        )

    ### if check_types works, this should raise an TypeError because j epxects a specific class and object is too general
    with pytest.raises(TypeError):
        checked_class(
            1,
            [[1]],
            needed_obj,
            1.0,
            {"a": 1},
            j=object,
        )

    ### if check_types works, this should raise an TypeError because j epxects a different class than needed_class
    with pytest.raises(TypeError):
        checked_class(
            1,
            [[1]],
            needed_obj,
            1.0,
            {"a": 1},
            j=needed_class,
        )

    ### if check_types works, this should raise an TypeError because j epxects a class not an int
    with pytest.raises(TypeError):
        checked_class(
            1,
            [[1]],
            needed_obj,
            1.0,
            {"a": 1},
            j=1,
        )

    ### if check_types works, this should raise an TypeError because g epxects a different class than needed_class
    with pytest.raises(TypeError):
        checked_class(
            1,
            [[1]],
            needed_obj,
            1.0,
            {"a": 1},
            g=needed_class,
        )

    ### if check_types works, this should raise an TypeError because h epxects a different class than needed_class
    with pytest.raises(TypeError):
        checked_class(
            1,
            [[1]],
            needed_obj,
            1.0,
            {"a": 1},
            h=needed_class,
        )


if __name__ == "__main__":
    test_check_types_decorator()



from typingchecker import check_types
import pytest


@check_types()
def tmp_test(dict_inside: dict[str, float], var_inside: float = 1.0):
    assert (
        isinstance(dict_inside["a"], float)
        and isinstance(dict_inside["b"], float)
        and isinstance(var_inside, float)
    ), "with float type hints, the types within the function should be converted to float"


@check_types(strictfloat=True)
def tmp_test_strict(dict_inside: dict[str, float], var_inside: float = 1.0):
    assert (
        isinstance(dict_inside["a"], float)
        and isinstance(dict_inside["b"], float)
        and isinstance(var_inside, float)
    ), "with float type hints, the types within the function should be converted to float"


def test_check_types_strictfloat():
    ####################################################################################
    ############################   RAISING NO ERRORS   #################################
    ####################################################################################

    ### does not raise an error
    ### because by default int values are converted to float if the type hint is float
    ### but the convertion to float does also affect the value outside of the function! (in case of dicts and lists)
    ### so if you have a list or dict containing int values and you do not want to unintentional convert them to float you can use strictfloat=True
    dict_outside = {"a": 1, "b": 2}
    tmp_test(dict_outside)
    assert isinstance(dict_outside["a"], float) and isinstance(
        dict_outside["b"], float
    ), "by default int values should be converted to float, this should also affect the value outside of the function!"

    ### for simple variables like b, the value outside of the function is not affected, but inside the function it's converted to float
    dict_outside = {"a": 1, "b": 2}
    var_outside = 2
    tmp_test(dict_outside, var_outside)
    assert isinstance(
        var_outside, int
    ), "the value outside of the function should not be affected by the convertion to float"

    ####################################################################################
    ##############################   RAISING ERRORS   ##################################
    ####################################################################################

    ### should raise an error because stricfloat=True and the values of "a" and "b" are an int instead of float
    dict_outside = {"a": 1, "b": 2}
    with pytest.raises(TypeError):
        tmp_test_strict(dict_outside)

    ### should also raise an error because stricfloat=True and here b is an int instead of float
    dict_outside = {"a": 1.0, "b": 2.0}
    with pytest.raises(TypeError):
        tmp_test_strict(dict_outside, 2)


if __name__ == "__main__":
    test_check_types_strictfloat()
